// This file is part of the `Selene` library.
// Copyright 2017-2018 Michael Hofmann (https://github.com/kmhofmann).
// Distributed under MIT license. See accompanying LICENSE file in the top-level directory.

#ifndef SELENE_IMG_PIXEL_PIXEL_TRAITS_HPP
#define SELENE_IMG_PIXEL_PIXEL_TRAITS_HPP

/// @file

#include <selene/base/Utils.hpp>

#include <selene/img/common/PixelFormat.hpp>

#include <selene/img/pixel/Pixel.hpp>

#include <cstdint>
#include <cstdlib>
#include <type_traits>

namespace sln {

/** \brief Class representing traits of a pixel.
 *
 * @tparam Element_ The pixel element type.
 */
template <typename Element_>
struct PixelTraits
{
  using Element = Element_;  ///< The pixel element type.
  static constexpr std::int16_t nr_channels = 1;  ///< The number of channels per pixel.
  static constexpr std::int16_t nr_bytes = sizeof(Element);  ///< The number of bytes per pixel.
  static constexpr std::int16_t nr_bytes_per_channel = sizeof(Element);  ///< The number of bytes per pixel channel.

  /// True, if the pixel elements are integral values; false otherwise.
  static constexpr bool is_integral = std::is_integral<Element>::value;

  /// True, if the pixel elements are floating point values; false otherwise.
  static constexpr bool is_floating_point = std::is_floating_point<Element>::value;

  /// True, if the pixel elements are unsigned; false otherwise.
  static constexpr bool is_unsigned = std::is_unsigned<Element>::value;

  /// The pixel format.
  static constexpr PixelFormat pixel_format = PixelFormat::Unknown;

  /// The sample format (unsigned/signed integer or floating point number).
  static constexpr SampleFormat sample_format =
      std::is_integral<Element>::value
          ? (std::is_unsigned<Element>::value ? SampleFormat::UnsignedInteger : SampleFormat::SignedInteger)
          : (std::is_floating_point<Element>::value ? SampleFormat::FloatingPoint : SampleFormat::Unknown);

  /// The value of the zero element.
  static constexpr Element zero_element = Element{0};
};

/** \brief Class representing traits of a pixel. Specialization for `Pixel<T, N, PF>`.
 *
 * @tparam T The channel element type of the pixel type.
 * @tparam N The number of channels of the pixel type.
 * @tparam PF The pixel format of the pixel type.
 */
template <typename T, std::size_t N, PixelFormat pixel_format_>
struct PixelTraits<Pixel<T, N, pixel_format_>>
{
  using Element = T;  ///< The pixel element type.
  static constexpr std::int16_t nr_channels = N;  ///< The number of channels per pixel.
  static constexpr std::int16_t nr_bytes = sizeof(Pixel<T, N, pixel_format_>);  ///< The number of bytes per pixel.
  static constexpr std::int16_t nr_bytes_per_channel = sizeof(Element);  ///< The number of bytes per pixel channel.

  /// True, if the pixel elements are integral values; false otherwise.
  static constexpr bool is_integral = std::is_integral<T>::value;

  /// True, if the pixel elements are floating point values; false otherwise.
  static constexpr bool is_floating_point = std::is_floating_point<T>::value;

  /// True, if the pixel elements are unsigned; false otherwise.
  static constexpr bool is_unsigned = std::is_unsigned<T>::value;

  /// The pixel format.
  static constexpr PixelFormat pixel_format = pixel_format_;

  /// The sample format (unsigned/signed integer or floating point number).
  static constexpr SampleFormat sample_format = std::is_integral<T>::value
                                                    ? (std::is_unsigned<T>::value ? SampleFormat::UnsignedInteger
                                                                                  : SampleFormat::SignedInteger)
                                                    : (std::is_floating_point<T>::value ? SampleFormat::FloatingPoint
                                                                                        : SampleFormat::Unknown);

  /// The value of the zero element.
  static constexpr Pixel<T, N, pixel_format_> zero_element =
    Pixel<T, N, pixel_format_>{make_array_n_equal<Element, N>(0)};
};

/** \brief Class representing traits of a pixel. Specialization for `const Pixel<T, N, PF>`.
 *
 * @tparam T The channel element type of the pixel type.
 * @tparam N The number of channels of the pixel type.
 * @tparam PF The pixel format of the pixel type.
 */
template <typename T, std::size_t N, PixelFormat pixel_format_>
struct PixelTraits<const Pixel<T, N, pixel_format_>>
{
  using Element = T;  ///< The pixel element type.
  static constexpr std::int16_t nr_channels = N;  ///< The number of channels per pixel.
  static constexpr std::int16_t nr_bytes = sizeof(const Pixel<T, N, pixel_format_>);  ///< The number of bytes per pixel.
  static constexpr std::int16_t nr_bytes_per_channel = sizeof(Element);  ///< The number of bytes per pixel channel.

  /// True, if the pixel elements are integral values; false otherwise.
  static constexpr bool is_integral = std::is_integral<T>::value;

  /// True, if the pixel elements are floating point values; false otherwise.
  static constexpr bool is_floating_point = std::is_floating_point<T>::value;

  /// True, if the pixel elements are unsigned; false otherwise.
  static constexpr bool is_unsigned = std::is_unsigned<T>::value;

  /// The pixel format.
  static constexpr PixelFormat pixel_format = pixel_format_;

  /// The sample format (unsigned/signed integer or floating point number).
  static constexpr SampleFormat sample_format = std::is_integral<T>::value
                                                    ? (std::is_unsigned<T>::value ? SampleFormat::UnsignedInteger
                                                                                  : SampleFormat::SignedInteger)
                                                    : (std::is_floating_point<T>::value ? SampleFormat::FloatingPoint
                                                                                        : SampleFormat::Unknown);

  /// The value of the zero element.
  static constexpr Pixel<T, N, pixel_format_> zero_element =
    Pixel<T, N, pixel_format_>{make_array_n_equal<Element, N>(0)};
};

// Out-of-line definitions for non-integral static declarations above:

template <typename Element_>
constexpr Element_ PixelTraits<Element_>::zero_element;

template <typename T, std::size_t N, PixelFormat pixel_format>
constexpr Pixel<T, N, pixel_format> PixelTraits<Pixel<T, N, pixel_format>>::zero_element;

template <typename T, std::size_t N, PixelFormat pixel_format>
constexpr Pixel<T, N, pixel_format> PixelTraits<const Pixel<T, N, pixel_format>>::zero_element;

}  // namespace sln

#endif  // SELENE_IMG_PIXEL_PIXEL_TRAITS_HPP
